// This test is intended to show the effects of LOCK CONTENTION.  We
// have several threads that repeatedly attempt to acquire a lock and
// update a shared counter. Because the code within the critical
// section is trivial, most of the time is spent trying to acquire the
// lock, and the parallel version of this code is slower than the
// sequential version.

// Intel quad-core: sequential 0.7s, parallel(N=4) 2s
// so the parallel code is 3x slower even though we are
// using four cores!

// Intel dual-core: sequential 0.9s, parallel(N=2) 1s
// so the parallel code is slightly slower even though we are
// using two cores!

// Of course, this is an extreme case, because our threads do not do
// anything time-consuming besides locking and unlocking. It does
// nevertheless illustrate the idea that "throwing more cores at the
// problem does not always help".

// A slightly different test would involve a longer critical section,
// say, updating a shared hash table. Because the code in the critical
// section is longer, lock contention is less important (I believe),
// and a lot of time is spent suspended. As a result, I would expect
// the parallel version of the code to be just as fast as the
// sequential version. It turns out that the parallel version is a bit
// faster; I suppose that the cost of locking is slightly parallelized.

import java.util.concurrent.locks.*;

class MyCThread implements Runnable {

  static long counter;
  static Lock lock = new ReentrantLock ();
    
  public void run ()
  {
    for (int i = 0; i < 10000000; i++) { // ten million
      lock.lock();
      counter++;
      lock.unlock();
    }
  }

}

public class LockContention {

  public static void test (boolean sequentially)
  {
    MyCThread.counter = 0;
    // The number of threads.
    final int N = 2;
    Thread[] thread = new Thread [N];
    for (int n = 0; n < N; n++)
      thread[n] = new Thread (new MyCThread ());
    try {
      long startTime = System.nanoTime();
      if (sequentially) {
	System.out.println("Running all threads in sequence...");
	for (int n = 0; n < N; n++) {
	  thread[n].start();
	  thread[n].join();
	}
      }
      else {
	System.out.println("Running all threads in parallel...");
	for (int n = 0; n < N; n++)
	  thread[n].start();
	for (int n = 0; n < N; n++)
	  thread[n].join();
      }
      long endTime = System.nanoTime();
      System.out.println("Done, the counter is " + MyCThread.counter);
      System.out.printf("Elapsed time: %.2fs\n", ((double) (endTime - startTime)) / 1000000000);
    } catch (InterruptedException e) {
      System.out.println("Interrupted.");
    }
  }

  public static void main (String[] args)
  {
    // Run the test several times, so the JIT has time to start up.
    for (int i = 0; i < 10; i++) {
      test(true);
      test(false);
      System.out.println();
    }
  }

}