// This test is intended to demonstrate LOCK OVERHEAD.

// The test is run sequentially in both cases (no parallelism).
// In one case, we don't use any locks; in the other case, we
// lock and unlock. So, we measure the cost of locking (in a
// situation where there is no contention at all).

// On my laptop, I find that the code is 100x slower with locks.
// This is not surprising, as the critical section is ridiculously
// short (incrementing a counter). If the critical section was a
// bit longer (e.g. updating a hash table), the ratio would be
// less extreme.

import java.util.concurrent.locks.*;

class MyOThread implements Runnable {

  boolean withLocks;
  static long counter;
  static Lock lock = new ReentrantLock ();

  public MyOThread (boolean withLocks)
  {
    this.withLocks = withLocks;
  }
    
  public void run ()
  {
    final int N = 10000000; // ten million
    if (withLocks)
      for (int i = 0; i < N; i++) {
	lock.lock();
	counter++;
	lock.unlock();
      }
    else
      for (int i = 0; i < N; i++)
	counter++;
  }

}

public class LockOverhead {

  public static void test (boolean withLocks)
  {
    MyOThread.counter = 0;
    // The number of threads.
    final int N = 2;
    Thread[] thread = new Thread [N];
    for (int n = 0; n < N; n++)
      thread[n] = new Thread (new MyOThread (withLocks));
    try {
      long startTime = System.nanoTime();
      // Run sequentially.
      for (int n = 0; n < N; n++) {
	thread[n].start();
	thread[n].join();
	}
      long endTime = System.nanoTime();
      System.out.println("Done, the counter is " + MyOThread.counter);
      System.out.printf("Elapsed time: %.2fs\n", ((double) (endTime - startTime)) / 1000000000);
    } catch (InterruptedException e) {
      System.out.println("Interrupted.");
    }
  }

  public static void main (String[] args)
  {
    // Run the test several times, so the JIT has time to start up.
    for (int i = 0; i < 10; i++) {
      System.out.println("Running sequentially, without lock/unlock...");
      test(false);
      System.out.println("Running sequentially, with lock/unlock...");
      test(true);
      System.out.println();
    }
  }

}